import zlib from 'zlib'
import http from 'http'
import https from 'https'


interface IRequestBodyUtil {
	json(data: any): Buffer
	urlencoded(data: any): Buffer
}

interface IRequestOption {
	url: string,
	method?: string
	body?: (util: IRequestBodyUtil) => Buffer
	headers?: Record<string, string | number>
	type?: 'json' | 'buffer' | 'text'
}

export function request<T = any>(option: IRequestOption) {
	return new Promise<T>((resolve, reject) => {

		const requestFunc = option.url.startsWith('https://') ? https.request : http.request
		const method = (option.method ?? 'get').toLowerCase()

		//处理body
		let contentType: string | null = null
		const body = option.body ? option.body({
			json(data) {
				contentType = 'application/json'
				return Buffer.from(JSON.stringify(data))
			},
			urlencoded(data) {
				contentType = 'application/x-www-form-urlencoded'
				return Buffer.from(Object.keys(data).map(k => `${k}=${encodeURIComponent(data[k])}`).join('&'))
			},
		}) : null

		//发起请求
		const req = requestFunc(option.url, {
			method,
			headers: {
				...option.headers,
				...(['get', 'options'].includes(method) || !body) ? {} : { 'Content-Length': body.byteLength },
				...contentType ? { 'Content-Type': contentType } : {},
			},
		}, res => {
			const buffer: Buffer[] = []
			res.on('data', d => buffer.push(d))
			res.on('error', reject)
			res.once('close', () => {
				let data = Buffer.concat(buffer)
				switch (res.headers['content-encoding']) {
					case 'gzip':
						data = zlib.gunzipSync(data)
						break
					case 'deflate':
						data = zlib.inflateSync(data)
						break
					case 'br':
						data = zlib.brotliDecompressSync(data)
						break
				}
				switch (option.type) {
					case 'text':
						resolve(data.toString() as any)
						break
					case 'json':
						resolve(JSON.parse(data.toString()))
						break
					default:
						resolve(data as any)
						break
				}
			})
		})
		req.on('error', reject)
		req.end(body)
	})
}
